import os
import re

from SphinxReport.Renderer import *
from SphinxReport.Tracker import *

# for trackers_derived_sets and trackers_master
if not os.path.exists("conf.py"):
    raise IOError( "could not find conf.py" )

execfile( "conf.py" )

def iterate_with_wrap( lines ):
    """iterate over lines in a Makefile

    The iterator merges multi-line lines.
    Comments before a variable are collected.
    """
    
    comments, buf = [], []

    keep = False

    for line in lines:
        if re.search( "Section parameters: start", line ): 
            keep = True
            comments = []
            continue
        if re.search( "Section parameters: end", line ): 
            break

        if line.startswith("#"):
            comments.append( re.sub( "^#+", "", line[:-1] ) )
            continue

        if not line.strip():
            comments = []
            continue

        if not keep: continue

        if line.endswith("\\\n"):
            buf.append( line[:-2].strip() )
            continue

        if buf:
            buf.append( line.strip() )
            yield " ".join( buf ), "\n".join(comments)
            buf = []            
            comments = []

        else:
            yield line, "\n".join(comments)
            comments = []

class MakefileParameters( Tracker ):
    """list parameters in the Makefile of a pipeline."""

    @returnLabeledData
    def __call__( self, track, slice = None ):

        rx = re.compile("^(\S[^=]*)\s*=\s*(.*)")
        infile = open( os.path.join( data_dir, track), "r" )
        data = []
        for line, comments in iterate_with_wrap( infile ):

            a = rx.search( line )
            if not a: continue
            
            k,v = a.groups() 
            
            if k.endswith("?"): k=k[:-1]
            if not v: v = "<empty>"
            if comments:
                data.append( (k, v + "\n\n" + comments ) )
            else:
                data.append( (k, v) )

        infile.close()
        return data



#!/usr/bin/env python
'''

    Script to interact with the existing conda installation.

    It assumes that a conda environment is active, and will
    run conda commands to:

    * Get the install folder for conda
    * Get a list of the existing environment
    * Pin all packages in all conda environments

    References:
    * https://conda.io/docs/user-guide/tasks/manage-pkgs.html#preventing-packages-from-updating-pinning
    * https://github.com/conda/conda/issues/7375

'''

import os
import subprocess
import json


def run_command(statement):
    os.environ.update({'BASH_ENV': os.path.join(os.environ['HOME'],'.bashrc')})
    process = subprocess.Popen(statement,
                               cwd=os.getcwd(),
                               shell=True,
                               executable='/bin/bash',
                               stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               env=os.environ.copy())

    (out, err) = process.communicate()

    if len(err) > 0:
        issue = (' Problem running command: {}\n'
                 ' Stderr was: {}\n'
                 .format(statement, err))
        raise Exception(issue)

    return out, err


statement = "conda info --json"
(out, err) = run_command(statement)
cmd_out = json.loads(out)
install_folder = cmd_out['conda_prefix']
#print(install_folder)

statement = "conda env list --json"
(out, err) = run_command(statement)
cmd_out = json.loads(out)
envs = cmd_out['envs']
#print(envs)

for env_folder in envs:
    if env_folder.startswith(install_folder) and \
       env_folder != install_folder:
        env_name = os.path.basename(env_folder)
        statement = "source activate {} && ".format(env_name)
        statement+= "conda list --json"
        (out, err) = run_command(statement)
        cmd_out = json.loads(out)
        conda_meta = install_folder + '/envs/' + env_name + '/conda-meta/pinned'
        with open(conda_meta, 'w') as outf:
            for pkg in cmd_out:
                outf.write(pkg['name'] + ' ==' + pkg['version'] + '\n')





1# setup.py
import sysconfig
import sys
import os
import subprocess
import re

# Import setuptools at the beginning
import setuptools
from setuptools import setup, find_packages, Extension
from distutils.version import LooseVersion
from Cython.Distutils import build_ext
from setuptools.command.build_py import build_py

# Ensure dependencies are installed before setup
try:
    import numpy
    import Cython
    import pysam
except ImportError as e:
    missing_package = str(e).split("'")[1]
    raise ImportError(f"{missing_package} must be installed before running setup.py")


class CustomBuildPy(build_py):
    def find_package_modules(self, package, package_dir):
        # Call the parent method
        modules = super().find_package_modules(package, package_dir)
        # Filter out `.py` directories
        filtered_modules = [
            module for module in modules
            if not os.path.isdir(os.path.join(package_dir, module[1] + '.py'))
        ]
        return filtered_modules

# Enforce Python 3 requirement
if sys.version_info < (3, 6):
    raise SystemExit("Python 3.6 or later is required to install this package.")

# Minimum setuptools version requirement
if LooseVersion(setuptools.__version__) < LooseVersion('1.1'):
    raise ImportError("Setuptools version >=1.1 is required")

# External dependency check
external_dependencies = [("wigToBigWig", "UCSC tools", 255), ("bedtools", "bedtools", 0)]
for tool, toolkit, expected in external_dependencies:
    retcode = subprocess.call(tool, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    if retcode != expected:
        print(f"WARNING: Dependency check for {toolkit} ({tool}) failed with error code {retcode}")

# Adjust packages and directories
cgat_packages = find_packages(where=".", include=["cgat", "cgat.*"], exclude=['tests', "tests.*"])
cgat_package_dirs = {'cgat': 'cgat'}
print("Discovered packages:", cgat_packages)

# Cython extensions and paths
conda_includes = [os.path.dirname(sysconfig.get_paths()["include"])]
conda_libdirs = [os.path.dirname(sysconfig.get_paths()["stdlib"])]
pysam_libraries = pysam.get_libraries()
pysam_libdirs = list(set(os.path.dirname(x) for x in pysam_libraries)) + conda_libdirs
pysam_libs = ["hts"] + [os.path.basename(x)[3:-3] for x in pysam_libraries]
pysam_dirname = os.path.dirname(pysam.__file__)
extra_link_args_pysam = [f'-Wl,-rpath,{x}' for x in pysam_libdirs + conda_libdirs]

extensions = [
    Extension(
        'cgat.Components',
        ['cgat/Components/Components.pyx', 'cgat/Components/connected_components.cpp'],
        include_dirs=[os.path.join('cgat', 'Components')] + conda_includes,
        language="c++",
    ),
    Extension(
        "cgat.NCL.cnestedlist",
        ["cgat/NCL/cnestedlist.pyx", "cgat/NCL/intervaldb.c"],
        include_dirs=["cgat/NCL", numpy.get_include()],  # Include NCL directory
        language="c",
    ),
    Extension(
        "cgat.GeneModelAnalysis",
        ["cgat/GeneModelAnalysis.pyx"],
        include_dirs=conda_includes + pysam.get_include() + [numpy.get_include()],
        define_macros=pysam.get_defines(),
        language="c",
    ),
    Extension(
        "cgat.BamTools.bamtools",
        ["cgat/BamTools/bamtools.pyx"],
        include_dirs=conda_includes + pysam.get_include() + [numpy.get_include()],
        library_dirs=pysam_libdirs,
        libraries=pysam_libs,
        define_macros=pysam.get_defines(),
        language="c",
        extra_link_args=extra_link_args_pysam,
    ),
    Extension(
        "cgat.BamTools.geneprofile",
        ["cgat/BamTools/geneprofile.pyx"],
        include_dirs=conda_includes + pysam.get_include() + [numpy.get_include()],
        library_dirs=pysam_libdirs,
        libraries=pysam_libs,
        define_macros=pysam.get_defines(),
        language="c",
        extra_link_args=extra_link_args_pysam,
    ),
    Extension(
        "cgat.BamTools.peakshape",
        ["cgat/BamTools/peakshape.pyx"],
        include_dirs=conda_includes + pysam.get_include() + [numpy.get_include()],
        library_dirs=pysam_libdirs,
        libraries=pysam_libs,
        define_macros=pysam.get_defines(),
        language="c",
        extra_link_args=extra_link_args_pysam,
    ),
    Extension(
        "cgat.VCFTools",
        ["cgat/VCFTools/vcftools.pyx"],
        include_dirs=conda_includes + pysam.get_include() + [numpy.get_include()],
        library_dirs=pysam_libdirs,
        libraries=pysam_libs,
        define_macros=pysam.get_defines(),
        language="c",
        extra_link_args=extra_link_args_pysam,
    ),
    Extension(
        "cgat.FastqTools",
        ["cgat/FastqTools/fastqtools.pyx"],
        include_dirs=conda_includes + pysam.get_include() + [numpy.get_include()],
        library_dirs=pysam_libdirs,
        libraries=pysam_libs,
        define_macros=pysam.get_defines(),
        language="c",
        extra_link_args=extra_link_args_pysam,
    ),
]

# Build setup configuration
setup(
    packages=cgat_packages,
    package_dir=cgat_package_dirs,
    include_package_data=True,
    package_data={
        "cgat.Components": ["*.h"],
    },
    ext_modules=extensions,
    cmdclass={'build_py': CustomBuildPy},
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'cgat = cgat.cgat.main',
            ],
        },
)



'''
cgat.py - Computational Genomics Analysis Tools
===============================================

:Tags: Genomics

To use a specific tool, type::

    cgat <tool> [tool options] [tool arguments]

Tools are grouped by keywords. For this message and a list of
available keywords type::

    cgat --help

For a list of tools matching a certain keyword, type::

   cgat --help <keyword>

or::

   cgat --help all

for a list of all available tools.

To get help for a specific tool, type::

    cgat <tool> --help
'''
import os
import sys
import re
import glob
import importlib.util
import collections
import cgatcore.iotools as iotools
import cgat


def mapKeyword2Script(path):
    '''Collect keywords from scripts.'''

    map_keyword2script = collections.defaultdict(list)

    for script in glob.glob(os.path.join(path, "*.py")):
        s = os.path.basename(script)[:-3]
        with iotools.open_file(script, 'r') as inf:
            data = [x for x in inf.readlines(10000) if x.startswith(':Tags:')]
            if data:
                keywords = [x.strip() for x in data[0][6:].split(' ')]
                for x in keywords:
                    if x:
                        map_keyword2script[x].append(s)

    return map_keyword2script


def printListInColumns(l, ncolumns):
    '''Output list *l* in *ncolumns*.'''
    ll = len(l)

    if ll == 0:
        return

    max_width = max([len(x) for x in l]) + 3
    n = ll // ncolumns
    if ll % ncolumns != 0:
        n += 1

    # Build columns
    columns = [l[x * n:x * n + n] for x in range(ncolumns)]

    # Add empty fields for missing columns in last row
    for x in range(ncolumns - (len(l) % ncolumns)):
        columns[-(x + 1)].append('')

    # Convert to rows
    rows = list(zip(*columns))

    # Build pattern for a row
    p = '%-' + str(max_width) + 's'
    pattern = ' '.join([p for _ in range(ncolumns)])

    # Put it all together
    return '\n'.join([pattern % row for row in rows])


def main(argv=None):
    argv = sys.argv

    path = os.path.join(os.path.abspath(os.path.dirname(cgat.__file__)),
                        "tools")

    if len(argv) == 1 or argv[1] in ("--help", "-h"):
        print(globals().get("__doc__", "No documentation available."))

        map_keyword2script = mapKeyword2Script(path)

        if len(argv) <= 2:
            print('cgat tools are grouped by keywords. The following keywords')
            print('are defined:\n')
            print("%s\n" % printListInColumns(list(map_keyword2script.keys()), 3))

        if 'all' in argv[2:]:
            print("The list of all available commands is:\n")
            print("%s\n" % printListInColumns(
                sorted([os.path.basename(x)[:-3]
                        for x in glob.glob(os.path.join(path, "*.py"))]),
                3))
        else:
            for arg in argv[2:]:
                if arg in map_keyword2script:
                    print("Tools matching the keyword '%s':\n" % arg)
                    print('%s\n' % printListInColumns(
                        sorted(map_keyword2script[arg]),
                        3))
        return

    if len(argv) < 2:
        print("Error: No command provided.\nUse --help for more information.", file=sys.stderr)
        sys.exit(1)

    command = argv[1]

    # Replace hyphens with underscores to match Python module naming conventions
    command = re.sub("-", "_", command)

    # Dynamically load the specified module using importlib
    module_path = os.path.join(path, f"{command}.py")

    if not os.path.isfile(module_path):
        print(f"Error: Module '{command}' not found in path '{path}'.", file=sys.stderr)
        sys.exit(1)

    spec = importlib.util.spec_from_file_location(command, module_path)
    if spec is None:
        print(f"Error: Cannot create a module spec for '{command}'.", file=sys.stderr)
        sys.exit(1)

    module = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(module)
    except Exception as e:
        print(f"Error: Failed to load module '{command}': {e}", file=sys.stderr)
        sys.exit(1)

    # Remove 'cgat' from sys.argv to pass the remaining arguments to the module's main function
    if len(sys.argv) > 0:
        del sys.argv[0]

    # Ensure the module has a 'main' function
    if not hasattr(module, 'main'):
        print(f"Error: The module '{command}' does not have a 'main' function.", file=sys.stderr)
        sys.exit(1)

    # Call the module's main function with the remaining arguments
    try:
        module.main(sys.argv)
    except Exception as e:
        print(f"Error: An exception occurred while executing the 'main' function of '{command}': {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    sys.exit(main())



"""collect scripts from a directory and create documentation
stubs.

Files starting with lower case characters are put into
the directory :file:`scripts`, while those starting with
upper case characters are put into :file:`modules`.
"""

TEMPLATE_SCRIPT='''
.. automodule:: %(prefix)s

.. program-output:: python ../scripts/%(prefix)s.py --help

'''

TEMPLATE_MODULE='''
.. automodule:: %(prefix)s
   :members:
   :show-inheritance:
'''

TEMPLATE_PIPELINE='''
.. automodule:: %(prefix)s
   :members:
   :show-inheritance:
'''

TEMPLATE_PIPELINEMODULE='''
.. automodule:: %(prefix)s
   :members:
   :show-inheritance:
'''

import glob
import os

import CGAT.experiment as E

if __name__ == "__main__":

    E.Start()

    dirs = ( ("../scripts/*.py", TEMPLATE_SCRIPT, 'scripts'),
             ("../CGAT/*.py", TEMPLATE_MODULE, 'modules'),
             ("../CGATPipelines/pipeline*.py", TEMPLATE_PIPELINE, 'pipelines'),
             ("../CGATPipelines/[A-Z]*.py", TEMPLATE_PIPELINEMODULE, 'pipelinemodules' ) )

    ncreated, nskipped = 0, 0

    for glob_expression, template, dest in dirs:

        if not os.path.exists( dest ):
            os.mkdir( dest )

        files = glob.glob( os.path.abspath( glob_expression ) )

        for filename in files:
            dirname, name = os.path.split( filename )
            prefix = name[:-3]

            #if os.path.exists( os.path.join( dirname, "_%s.pyx" % prefix )):
            #    E.warn( "ignoring pyximport file _%s.pyx" % prefix )
            #    continue

            filename = os.path.join( os.path.abspath(dest), "%s.rst" % prefix )
            if os.path.exists( filename ):
                nskipped += 1
                continue

            E.debug( "adding %s" % filename )
            outfile = open( filename, "w" )
            outfile.write( template % locals() )
            outfile.close()

            ncreated += 1

    E.info( "ncreated=%i, nskipped=%i" % (ncreated, nskipped ) )
    
    E.Stop()



'''
cgat_check_deps.py - check whether the software dependencies are on your PATH
=============================================================================

Purpose
-------

.. The goal of this script is to provide a list of third-party command-line
programs used in a Python script given as input, and check whether
they are on your PATH. This is useful to identify dependencies across all
CGAT pipelines and module files.

This script takes the path to a Python script, which is expected to call
command-line programs like we do in CGAT pipelines, i.e.:

   statement = """cmd-1 | cmd-2 | cmd-3""""
   P.run()

Programs called other way (e.g. using subprocess) will not be picked up
by this script.

Usage
-----

.. python cgat_check_deps --pipeline </path/to/pipeline_name.py> [--print-summary]

Example::

   python cgat_check_deps --pipeline CGATPipelines/pipeline_annotations.py

Type::

   cgat cgat_check_deps --help

for command line help.

Command line options
--------------------

'''

import os
import shutil
import sys
import re
import ast
import argparse
import bashlex


# inspired by
# https://docs.python.org/3/library/ast.html#module-ast
# http://bit.ly/2rDf5xu
# http://bit.ly/2r0Uv9t
# really helpful, used astviewer (installed in a conda-env) to inspect examples
# https://github.com/titusjan/astviewer
def is_cgat_statement(node):
    '''
       Auxiliary function to check for cgat statement:
           statement = "command"
    '''

    result = False
    result = type(node) is ast.Assign and \
        hasattr(node, 'targets') and \
        hasattr(node.targets[0], 'id') and \
        node.targets[0].id == "statement"

    return result


def is_cgat_executable(node):
    '''
       Auxiliary function to check for cgat statement:
           executable = "command"
    '''

    result = False
    result = type(node) is ast.Assign and \
        hasattr(node, 'targets') and \
        hasattr(node.targets[0], 'id') and \
        node.targets[0].id == "executable"

    return result


def is_cgat_executable_name(node):
    '''
       Auxiliary function to check for cgat statement:
           executable_name = "command"
    '''

    result = False
    result = type(node) is ast.Assign and \
        hasattr(node, 'targets') and \
        hasattr(node.targets[0], 'id') and \
        node.targets[0].id == "executable_name"

    return result


def is_cgat_append(node):
    '''
       Auxiliary function to check for cgat statement:
           statment.append("command")
    '''

    result = False
    result = type(node) is ast.Expr and \
        hasattr(node, 'value') and \
        hasattr(node.value, 'func') and \
        hasattr(node.value.func, 'value') and \
        hasattr(node.value.func.value, 'id') and \
        node.value.func.value.id == "statement" and \
        hasattr(node.value.func, 'attr') and \
        node.value.func.attr == "append"

    return result


def get_cmd_string(node):
    '''
       Auxiliary function to get commands in the cgat statement:
           statement = "command"
    '''

    result = ""
    if hasattr(node.value, 's'):
        result = node.value.s
    elif hasattr(node.value, 'left') and hasattr(node.value.left, 's'):
        result = node.value.left.s

    return result


def get_append_string(node):
    '''
       Auxiliary function to get commands in the cgat statement:
           statement.append("command")
    '''

    result = ""
    if hasattr(node, 'value') and \
       hasattr(node.value, 'args') and \
       hasattr(node.value.args[0], 's'):
        result = node.value.args[0].s
    elif hasattr(node, 'value') and \
            hasattr(node.value, 'args') and \
            hasattr(node.value.args[0], 'left') and \
            hasattr(node.value.args[0].left, 's'):
        result = node.value.args[0].left.s

    return result


def cleanup_statement(statement):
    '''
       Auxiliary function to cleanup cgat statements
    '''
    # cleanup whitespaces, tabs, and newlines
    result = " ".join(statement.split())
    # cleanup parameter interpolation
    result = re.sub("\%\(\w+\)\w+", "CGATparameter", result)
    return result


# Thanks to: https://github.com/idank/bashlex
# Workflow:
# statement = ''' <whatever> '''
# statement = " ".join(statement.split())
# statement = re.sub("\%\(", "", statement)
# statement = re.sub("\)s", "", statement)
# parts = bashlex.parse(statement)
# commands = []
# get_cmd_names(parts[0], commands)
def get_cmd_names(tree, commands):
    if hasattr(tree, 'parts') and len(tree.parts) == 0:
        return
    else:
        if hasattr(tree, 'kind'):
            if tree.kind == 'command' and hasattr(tree.parts[0], 'word'):
                commands.append(str(tree.parts[0].word))
            if (tree.kind == 'processsubstitution' or tree.kind == 'commandsubstitution') and \
                    hasattr(tree, 'command') and hasattr(tree.command, 'parts') and \
                    hasattr(tree.command.parts[0], 'word'):
                commands.append(str(tree.command.parts[0].word))
        if hasattr(tree, 'parts'):
            for e in tree.parts:
                get_cmd_names(e, commands)
        if hasattr(tree, 'command'):
            for e in tree.command.parts:
                get_cmd_names(e, commands)


def checkDepedencies(pipeline):

    # check existence of pipeline script
    if not os.access(pipeline, os.R_OK):
        raise IOError("Pipeline %s was not found\n" % pipeline)

    if os.path.isdir(pipeline):
        raise IOError("The given input is a folder, and must be a script\n")

    # parse pipeline script
    with open(pipeline) as f:
        tree = ast.parse(f.read())

    # list to store all statements = ''' <commands> '''
    statements = []

    # inspired by
    # https://docs.python.org/3/library/ast.html#module-ast
    # http://bit.ly/2rDf5xu
    # http://bit.ly/2r0Uv9t
    # really helpful, used astviewer (installed in a conda-env) to inspect examples
    # https://github.com/titusjan/astviewer
    for node in ast.walk(tree):
        statement = ""
        if is_cgat_statement(node) or \
           is_cgat_executable(node) or \
           is_cgat_executable_name(node):

            statement = get_cmd_string(node)

        elif is_cgat_append(node):
            statement = get_append_string(node)

        if len(statement) > 0 and not statement.startswith(' -'):
            #print(statement)
            statement = cleanup_statement(statement)
            statements.append(statement)

    # dictionary where:
    # key = program name
    # value = number of times it has been called
    deps = {}

    # set of names that are not proper deps
    exceptions = ['create',
                  'drop',
                  'select',
                  'attach',
                  'insert',
                  'module',
                  '%s',
                  'tool',
                  'cmd-farm',
                  'cmd-sql',
                  'cmd_extract',
                  'cmds',
                  'compress',
                  'conda_env',
                  'filter_cmd',
                  'load_statement',
                  'match_executable',
                  'rscript',
                  'paired_processing',
                  'executable',
                  'transform',
                  'uncompress',
                  'execglam2',
                  'execglam2scan',
                  'CGATparameter',
                  'checkpoint',
                  'for']

    for statement in statements:
        # use bashlex to parse statements
        commands = []
        try:
            #print(statement)
            parts = bashlex.parse(statement)
            get_cmd_names(parts[0], commands)
        except bashlex.errors.ParsingError:
            pass

        for command in commands:
            #print(command)
            if command.lower() not in exceptions:
                if command not in deps:
                    deps[command] = 1
                else:
                    deps[command] += 1

    # list of unmet dependencies
    check_path_failures = []

    # print dictionary ordered by value
    for k in sorted(deps, key=deps.get, reverse=True):
        if shutil.which(k) is None:
            check_path_failures.append(k)

    return deps, check_path_failures


def main(argv=None):
    """script main.
    parses command line options in sys.argv, unless *argv* is given.
    """

    if (sys.version_info < (3, 0, 0)):
        raise OSError("This script is Python 3 only")
        sys.exit(-1)

    if argv is None:
        argv = sys.argv

    # setup command line parser
    parser = argparse.ArgumentParser(description='Get 3rd party dependencies.')

    parser.add_argument("pipeline", help="Path to CGAT pipeline or module")

    parser.add_argument("-s", "--print-summary", dest="summary",
                        action="store_true", default=False,
                        help="Print how many times a program is used")

    options = parser.parse_args()

    # get dependencies dependencies
    deps, check_path_failures = checkDepedencies(options.pipeline)

    # print info about dependencies
    if len(deps) == 0:
        print('\nNo dependencies found.\n')
    else:
        # print dictionary ordered by value
        if options.summary:
            for k in sorted(deps, key=deps.get, reverse=True):
                print('\nProgram: {0!s} used {1} time(s)'.format(k, deps[k]))

        n_failures = len(check_path_failures)
        if n_failures == 0:
            print('\nCongratulations! All required programs are available on your PATH\n')
        else:
            print('\nThe following programs are not on your PATH')
            for p in check_path_failures:
                print('\n{0!s}'.format(p))
            print


if __name__ == "__main__":
    sys.exit(main(sys.argv))
